"""Implements functionality unique to the Lake Shore M81."""
from datetime import datetime
import struct
from base64 import b64decode
from threading import Lock
from warnings import warn

from lakeshore.ssm_system_enums import SSMSystemEnums
from lakeshore.xip_instrument import XIPInstrument, XIPInstrumentException, RegisterBase
from lakeshore.ssm_measure_module import MeasureModule
from lakeshore.ssm_source_module import SourceModule
from lakeshore.ssm_settings_profiles import SettingsProfiles
from lakeshore.requires_firmware_version import requires_firmware_version

try:
    from wakepy import keepawake
except NotImplementedError:
    pass  # Proceed without wakepy on linux without systemd
except KeyError:
    pass  # Proceed without wakepy on linux without dbus


class SSMSystemOperationRegister(RegisterBase):
    """Class object representing the operation status register"""

    bit_names = [
        "s1_summary",
        "s2_summary",
        "s3_summary",
        "m1_summary",
        "m2_summary",
        "m3_summary",
        "data_stream_in_progress"
    ]

    def __init__(self,
                 s1_summary,
                 s2_summary,
                 s3_summary,
                 m1_summary,
                 m2_summary,
                 m3_summary,
                 data_stream_in_progress):
        self.s1_summary = s1_summary
        self.s2_summary = s2_summary
        self.s3_summary = s3_summary
        self.m1_summary = m1_summary
        self.m2_summary = m2_summary
        self.m3_summary = m3_summary
        self.data_stream_in_progress = data_stream_in_progress


class SSMSystemQuestionableRegister(RegisterBase):
    """Class object representing the questionable status register"""

    bit_names = [
        "s1_summary",
        "s2_summary",
        "s3_summary",
        "m1_summary",
        "m2_summary",
        "m3_summary",
        "critical_startup_error",
        "critical_runtime_error",
        "heartbeat",
        "calibration",
        "data_stream_overflow"
    ]

    # pylint: disable=too-many-arguments
    def __init__(self,
                 s1_summary,
                 s2_summary,
                 s3_summary,
                 m1_summary,
                 m2_summary,
                 m3_summary,
                 critical_startup_error,
                 critical_runtime_error,
                 heartbeat,
                 calibration,
                 data_stream_overflow):
        self.s1_summary = s1_summary
        self.s2_summary = s2_summary
        self.s3_summary = s3_summary
        self.m1_summary = m1_summary
        self.m2_summary = m2_summary
        self.m3_summary = m3_summary
        self.critical_startup_error = critical_startup_error
        self.critical_runtime_error = critical_runtime_error
        self.heartbeat = heartbeat
        self.calibration = calibration
        self.data_stream_overflow = data_stream_overflow


class SSMSystem(XIPInstrument, SSMSystemEnums):
    """Class for interaction with the M81 instrument"""

    vid_pid = [(0x1FB9, 0x0704), (0x10C4, 0xEA60)]

    def __init__(self,
                 serial_number=None,
                 com_port=None,
                 baud_rate=921600,
                 flow_control=True,
                 timeout=2.0,
                 ip_address=None,
                 tcp_port=7777,
                 **kwargs):

        # Call the parent init, then fill in values specific to SSM
        XIPInstrument.__init__(self,
                               serial_number,
                               com_port,
                               baud_rate,
                               flow_control,
                               timeout,
                               ip_address,
                               tcp_port,
                               **kwargs)

        self.operation_register = SSMSystemOperationRegister
        self.questionable_register = SSMSystemQuestionableRegister

        # Instantiate modules
        self.source_modules = [SourceModule(i + 1, self) for i in range(self.get_num_source_channels())]
        self.measure_modules = [MeasureModule(i + 1, self) for i in range(self.get_num_measure_channels())]

        self.settings_profiles = SettingsProfiles(self)

        self.stream_lock = Lock()

        # Sweeping limits
        self.min_sweep_dwell = 0.000_2
        self.max_sweep_points = 100_001

    def load_modules(self):
        """Loads all unloaded modules. Connected modules must be loaded before they can be used"""
        self.command('SYSTem:LOAD')

    def get_num_measure_channels(self):
        """Returns the number of measure channels supported by the instrument"""

        return int(self.query('SENSe:NCHannels?'))

    def get_num_source_channels(self):
        """Returns the number of source channels supported by the instrument"""

        return int(self.query('SOURce:NCHannels?'))

    def get_source_module(self, port_number):
        """Returns a SourceModule instance for the given port number"""

        try:
            return self.source_modules[port_number - 1]
        except IndexError:
            raise IndexError(
                f'Invalid port number. Must be between 1 and {self.get_num_source_channels()}') from None

    def get_source_pod(self, port_number):
        """alias of get_source_module"""
        return self.get_source_module(port_number)

    def get_source_module_by_name(self, module_name):
        """Return the SourceModule instance that matches the specified name"""

        return self._locate_module_by_name(module_name, self.source_modules)

    def get_measure_module(self, port_number):
        """Returns a MeasureModule instance for the given port number"""

        try:
            return self.measure_modules[port_number - 1]
        except IndexError:
            raise IndexError(
                f'Invalid port number. Must be between 1 and {self.get_num_measure_channels()}') from None

    def get_measure_pod(self, port_number):
        """alias of get_measure_module"""
        return self.get_measure_module(port_number)

    def get_measure_module_by_name(self, module_name):
        """Return the MeasureModule instance that matches the specified name"""

        return self._locate_module_by_name(module_name, self.measure_modules)

    @staticmethod
    def _locate_module_by_name(module_name, set_of_modules):
        module_names = []
        for module in set_of_modules:
            try:
                module_names.append(module.get_name())
            except XIPInstrumentException as exception:
                if '-241,"Hardware missing;' not in str(exception):
                    raise
                module_names.append(None)

        num_matches = len([name for name in module_names if name == module_name])

        if num_matches < 1:
            raise XIPInstrumentException(f'No module was found with the name {module_name}.')
        if num_matches > 1:
            raise XIPInstrumentException(f'Module name conflict: more than one module is named {module_name}.')

        return set_of_modules[module_names.index(module_name)]

    data_source_types = {
        'RTIMe': float,
        'SAMPlitude': float,
        'SOFFset': float,
        'SFRequency': float,
        'SRANge': float,
        'SVLimit': lambda s: bool(int(s)),
        'SILimit': lambda s: bool(int(s)),
        'MDC': float,
        'MRMs': float,
        'MPPeak': float,
        'MNPeak': float,
        'MPTPeak': float,
        'MX': float,
        'MY': float,
        'MR': float,
        'MTHeta': float,
        'MRANge': float,
        'MOVerload': lambda s: bool(int(s)),
        'MSETtling': lambda s: bool(int(s)),
        'MUNLock': lambda s: bool(int(s)),
        'MRFRequency': float,
        'GPIStates': int,
        'GPOStates': int,
        'SRSettling': lambda s: bool(int(s)),
    }

    data_source_lookup = {}
    for mnemonic, channel_index in data_source_types.items():
        short_form = ''.join(c for c in mnemonic if not c.islower())
        data_source_lookup[mnemonic.upper()] = channel_index
        data_source_lookup[short_form.upper()] = channel_index

    def get_multiple(self, *data_sources):
        """
        This function is deprecated. Use fetch_multiple() instead.

        .. deprecated:: 1.5.4
            Use fetch_multiple instead.
        """

        warn("The get_multiple method is deprecated, use fetch_multiple instead", DeprecationWarning)
        return self.fetch_multiple(*data_sources)

    def get_multiple_min_max_values(self, *data_sources):
        """Gets a synchronized minimum and maximum value for each specified data source.

            Args:
                data_sources (str, int): Pairs of (DATASOURCE_MNEMONIC, CHANNEL_INDEX).
        """

        elements = ','.join(f'{mnemonic},{index}' for (mnemonic, index) in data_sources)
        response_values = self.query(f'STAT:MMAX? {elements}').split(',')

        return [(float(response_values[i]), float(response_values[i + 1])) for i in range(0, len(response_values), 2)]

    def stream_data(self, rate, num_points, *data_sources):
        """Generator object to stream data from the instrument.

            Args:
                rate (int): Desired transfer rate in points/sec.
                num_points (int): Number of points to return. None to stream indefinitely.

                data_sources (SSMSystemDataSourceMnemonic or str, int):
                    Variable length list of pairs of (DATA_SOURCE, CHANNEL_INDEX).

            Yields:
                A single row of stream data as a tuple
        """

        with self.stream_lock:
            with keepawake(keep_screen_awake=True):
                self.command('TRACe:RESEt')
                self._configure_stream_elements(data_sources)
                self.command('TRACe:FORMat:ENCOding B64')
                self.command(f'TRACe:RATE {rate}')

                bytes_per_row = int(self.query('TRACe:FORMat:ENCOding:B64:BCOunt?'))
                binary_format = '<' + self.query('TRACe:FORMat:ENCOding:B64:BFORmat?').strip('\"')

                if num_points is not None:
                    self.command(f'TRACe:STARt {num_points}')
                else:
                    self.command('TRACe:STARt')

                num_collected = 0
                while num_points is None or num_collected < num_points:
                    b64_string = ''
                    while not b64_string:
                        b64_string = self.query('TRACe:DATA:ALL?', check_errors=False)

                    new_bytes = b64decode(b64_string)
                    rows = [new_bytes[i:i + bytes_per_row] for i in range(0, len(new_bytes), bytes_per_row)]

                    for row in rows:
                        data = struct.unpack(binary_format, row)
                        num_collected += 1

                        yield data

            overflow_occurred = bool(int(self.query('TRACe:DATA:OVERflow?', check_errors=True)))
            if overflow_occurred:
                raise XIPInstrumentException('Data loss occurred during this data stream.')

    def get_data(self, rate, num_points, *data_sources):
        """Like stream_data, but returns a list.

            Args:
                rate (int): Desired transfer rate in points/sec.
                num_points (int): Number of points to return.

                data_sources (SSMSystemDataSourceMnemonic or str, int):
                    Variable length list of pairs of (DATA_SOURCE, CHANNEL_INDEX).

            Returns:
                All available stream data as a list of tuples
        """

        return list(self.stream_data(rate, num_points, *data_sources))

    def log_data_to_csv_file(self, rate, num_points, file, *data_sources, **kwargs):
        """Like stream_data, but logs directly to a CSV file.

            Args:
                rate (int): Desired transfer rate in points/sec.
                file (IO): File to log to.
                num_points (int): Number of points to log.
                data_sources (SSMSystemDataSourceMnemonic or str, int): Pairs of (DATA_SOURCE, CHANNEL_INDEX).
                write_header (bool): If true, a header row is written with column names.
        """
        write_header = kwargs.pop('write_header', True)

        if write_header:
            self._configure_stream_elements(data_sources)
            header = self.query('TRACe:FORMat:HEADer?').strip('\"')
            file.write(header + '\n')

        for row in self.stream_data(rate, num_points, *data_sources):
            file.write(','.join(str(x) for x in row) + '\n')

    def _configure_stream_elements(self, data_sources):
        elements = ','.join(f'{mnemonic},{index}' for (mnemonic, index) in data_sources)
        self.command(f'TRACe:FORMat:ELEMents {elements}')

    def get_ref_in_edge(self):
        """Returns the active edge of the reference input. 'RISing' or 'FALLing'."""

        return self.query('INPut:REFerence:EDGe?')

    def set_ref_in_edge(self, edge):
        """Sets the active edge of the reference input

            Args:
                edge (str):
                    The new active edge ('RISing', or 'FALLing')
        """

        self.command(f'INPut:REFerence:EDGe {edge}')

    def get_ref_out_source(self):
        """Returns the channel used for the reference output. 'S1', 'S2', or 'S3'."""

        return self.query('OUTPut:REFerence:SOURce?')

    def set_ref_out_source(self, ref_out_source):
        """Sets the channel used for the reference output.

            Args:
                ref_out_source (str):
                    The new reference out source ('S1', 'S2', or 'S3')
        """

        self.command(f'OUTPut:REFerence:SOURce {ref_out_source}')

    def get_ref_out_state(self):
        """Returns the enable state of reference out"""

        return bool(int(self.query('OUTPut:REFerence:STATe?')))

    def set_ref_out_state(self, ref_out_state):
        """Sets the enable state of reference out

            Args:
                ref_out_state (bool):
                    The new reference out state (True to enable reference out, False to disable reference out)
        """

        self.command(f'OUTPut:REFerence:STATe {str(int(ref_out_state))}')

    def enable_ref_out(self):
        """Sets the enable state of reference out to True"""

        self.set_ref_out_state(True)

    def disable_ref_out(self):
        """Sets the enable state of reference out to False"""

        self.set_ref_out_state(False)

    def configure_ref_out(self, ref_out_source, ref_out_state=True):
        """Configure the reference output

            Args:
                ref_out_source (str):
                    The new reference out source ('S1', 'S2', or 'S3')

                ref_out_state (bool):
                    The new reference out state (True to enable reference out, False to disable reference out)
        """

        self.set_ref_out_source(ref_out_source)
        self.set_ref_out_state(ref_out_state)

    def get_mon_out_mode(self):
        """Returns the channel used for the monitor output. 'M1', 'M2', 'M3', or 'MANUAL'."""

        return self.query('OUTPut:MONitor:MODe?')

    def set_mon_out_mode(self, mon_out_source):
        """Sets the channel used for the monitor output.

            Args:
                mon_out_source (str):
                    The new monitor out source ('M1', 'M2', 'M3', or 'MANUAL')
        """

        self.command(f'OUTPut:MONitor:MODe {mon_out_source}')

    def get_mon_out_state(self):
        """Returns the enable state of monitor out"""

        return bool(int(self.query('OUTPut:MONitor:STATe?')))

    def set_mon_out_state(self, mon_out_state):
        """Sets the enable state of monitor out

            Args:
                mon_out_state (bool):
                    The new monitor out state (True to enable monitor out, False to disable monitor out)
        """

        self.command(f'OUTPut:MONitor:STATe {str(int(mon_out_state))}')

    def enable_mon_out(self):
        """Sets the enable state of monitor out to True"""

        self.set_mon_out_state(True)

    def disable_mon_out(self):
        """Sets the enable state of monitor out to False"""

        self.set_mon_out_state(False)

    def configure_mon_out(self, mon_out_source, mon_out_state=True):
        """Configure the monitor output

            Args:
                mon_out_source (str):
                    The new monitor out source ('M1', 'M2', or 'M3')

                mon_out_state (bool):
                    The new monitor out state (True to enable monitor out, False to disable monitor out)
        """

        self.set_mon_out_mode(mon_out_source)
        self.set_mon_out_state(mon_out_state)

    def get_mon_out_scale(self):
        """Returns the monitor out scaling factor of the configured module."""

        return float(self.query('OUTPut:MONitor:SCALe?'))

    def get_head_cal_datetime(self):
        """Returns the date and time of the head calibration"""

        response = self.query('CALibration:DATE?').split(',')
        return datetime(int(response[0]), int(response[1]), int(response[2]), int(response[3]), int(response[4]), int(response[5]))

    def get_head_cal_temperature(self):
        """Returns the temperature of the head calibration"""

        return float(self.query('CALibration:TEMPerature?'))

    def get_head_self_cal_status(self):
        """Returns the status of the last head self calibration"""

        return self.query('CALibration:SCALibration:STATus?')

    def get_head_self_cal_datetime(self):
        """Returns the datetime of the last head self calibration"""

        response = self.query('CALibration:SCALibration:DATE?').split(',')
        return datetime(int(response[0]), int(response[1]), int(response[2]), int(response[3]), int(response[4]), int(response[5]))

    def get_head_self_cal_temperature(self):
        """Returns the temperature of the last head self calibration"""

        return float(self.query('CALibration:SCALibration:TEMPerature?'))

    def run_head_self_calibration(self):
        """"Runs a self calibration for the head"""

        self.command('CALibration:SCALibration:RUN')

    def reset_head_self_calibration(self):
        """"Restore the factory self calibration"""

        self.command('CALibration:SCALibration:RESet')

    def set_mon_out_manual_level(self, manual_level):
        """Set the manual level of monitor out when the mode is MANUAL

            Args:
                manual_level (float):
                    The new monitor out manual level
        """

        self.command(f'OUTPut:MONitor:MLEVel {str(manual_level)}')

    def get_mon_out_manual_level(self):
        """Returns the manual level of monitor out"""

        return float(self.query('OUTPut:MONitor:MLEVel?'))

    def configure_mon_out_manual_mode(self, manual_level, mon_out_state=True):
        """Configures the monitor output for manual mode

            Args:
                manual_level (float):
                    The new monitor out manual level

                mon_out_state (bool):
                    The new monitor out state (True to enable monitor out, False to disable monitor out)
        """

        self.set_mon_out_manual_level(manual_level)
        self.set_mon_out_mode('MANUAL')
        self.set_mon_out_state(mon_out_state)

    def get_line_frequency(self):
        """Returns the line frequency in Hz"""

        return float(self.query('SYSTem:LFRequency?'))

    def get_detected_line_frequency(self):
        """Returns the detected line frequency in Hz"""

        return float(self.query('SYSTem:LFRequency:DETected?'))

    def get_line_frequency_detection_error_status(self):
        """Returns the line frequency detection error status. True if the frequency is out of bounds."""

        return bool(int(self.query('SYSTem:LFRequency:ERRor?')))

    def fetch_multiple(self, *data_sources):
        """Gets a list of the latest values corresponding to the input data sources, and returns them as soon as
        possible.

            Args:
                data_sources (SSMSystemDataSourceMnemonic or str, int):
                    Variable length list of pairs of (DATA_SOURCE, CHANNEL_INDEX).

            Returns:
                Tuple of values corresponding to the given data sources
        """

        elements = ','.join(f'{mnemonic},{index}' for (mnemonic, index) in data_sources)
        response_values_with_indices = enumerate(self.query(f'FETCh? {elements}').split(','))

        return tuple(
            (self.data_source_lookup[data_sources[i][0].upper()])(value) for (i, value) in response_values_with_indices)

    def read_multiple(self, *data_sources):
        """Initiates measurement of new values corresponding to the input data sources, and returns them after the
        measurement is complete.

            Args:
                data_sources (SSMSystemReadDataSourceMnemonic or str, int):
                    Variable length list of pairs of (DATA_SOURCE, CHANNEL_INDEX).

            Returns:
                Tuple of values corresponding to the given data sources
        """

        elements = ','.join(f'{mnemonic},{index}' for (mnemonic, index) in data_sources)
        response_values_with_indices = enumerate(self.query(f'READ? {elements}').split(','))

        return tuple(
            (self.data_source_lookup[data_sources[i][0].upper()])(value) for (i, value) in response_values_with_indices)

    @requires_firmware_version('1.7.0')
    def initiate_sweeps(self):
        """Initiates sweeps across all channels."""

        self.command('SWEep:INITiate')

    @requires_firmware_version('1.7.0')
    def abort_sweeps(self):
        """Aborts in progress sweeps across all channels."""

        self.command('SWEep:ABORt')



"""Contains enumerations specific to the M81 """

from enum import Enum


class SSMSystemEnums:
    """Class for collecting the enumerations specific to the SSMSystem without bulking up that class size"""

    class DataSourceMnemonic(str, Enum):
        """Enumeration of M81 data source mnemonics"""
        RELATIVE_TIME = 'RTIMe'
        SOURCE_AMPLITUDE = 'SAMPlitude'
        SOURCE_OFFSET = 'SOFFset'
        SOURCE_FREQUENCY = 'SFRequency'
        SOURCE_RANGE = 'SRANge'
        SOURCE_VOLTAGE_LIMIT = 'SVLimit'
        SOURCE_CURRENT_LIMIT = 'SILimit'
        SOURCE_IS_SWEEPING = 'SSWeeping'
        MEASURE_DC = 'MDC'
        MEASURE_RMS = 'MRMS'
        MEASURE_POSITIVE_PEAK = 'MPPeak'
        MEASURE_NEGATIVE_PEAK = 'MNPeak'
        MEASURE_PEAK_TO_PEAK = 'MPTPeak'
        MEASURE_X = 'MX'
        MEASURE_Y = 'MY'
        MEASURE_R = 'MR'
        MEASURE_THETA = 'MTHeta'
        MEASURE_RANGE = 'MRANge'
        MEASURE_OVERLOAD = 'MOVerload'
        MEASURE_SETTLING = 'MSETtling'
        MEASURE_UNLOCK = 'MUNLock'
        MEASURE_REFERENCE_FREQUENCY = 'MRFRequency'
        GENERAL_PURPOSE_INPUT_STATES = 'GPIStates'
        GENERAL_PURPOSE_OUTPUT_STATES = 'GPOStates'

        # Gets around having to use .value to access the string
        def __str__(self) -> str:
            return str.__str__(self)

    class ReadDataSourceMnemonic(str, Enum):
        """Enumeration of M81 read data source mnemonics"""
        MEASURE_DC = 'MDC'
        MEASURE_RMS = 'MRMs'
        MEASURE_POSITIVE_PEAK = 'MPPeak'
        MEASURE_NEGATIVE_PEAK = 'MNPeak'
        MEASURE_PEAK_TO_PEAK = 'MPTPeak'
        MEASURE_RANGE = 'MRANge'

        # Gets around having to use .value to access the string
        def __str__(self) -> str:
            return str.__str__(self)

    class ExcitationType(str, Enum):
        """Class object representing the possible excitation types of a source module."""
        CURRENT = 'CURRENT'
        VOLTAGE = 'VOLTAGE'

        # Gets around having to use .value to access the string
        def __str__(self) -> str:
            return str.__str__(self)

    class SourceSweepType(str, Enum):
        """Class representing the available sweep types for a source module."""
        CURRENT_AMPLITUDE = 'CURRent'
        VOLTAGE_AMPLITUDE = 'VOLTage'

        # Gets around having to use .value to access the string
        def __str__(self) -> str:
            return str.__str__(self)

    class SourceSweepSettings:
        """Class to configure a parameter sweep on a source module."""

        class SweepSpacing(str, Enum):
            """Class object representing the possible types of sweep spacing."""
            LINEAR = 'LINEAR'
            LOGARITHMIC = 'LOGARITHMIC'

            # Gets around having to use .value to access the string
            def __str__(self) -> str:
                return str.__str__(self)

        class Direction(str, Enum):
            """Class object representing the possible directions for sweeping"""
            DOWN = 'DOWN'
            UP = 'UP'

            # Gets around having to use .value to access the string
            def __str__(self) -> str:
                return str.__str__(self)

        def __init__(self, sweep_type, start, stop, points, dwell,
                     direction=Direction.UP, round_trip=False, spacing=SweepSpacing.LINEAR):
            """
            Constructor for SourceModuleSweepSettings class

            Args:
                sweep_type (SourceSweepType):
                    The type of sweep to perform.
                start (float):
                    Sets the start value of the source sweep.
                stop (float):
                    Sets the stop value of the source sweep.
                points (int):
                    Sets the number of steps in the source sweep.
                dwell (float):
                    Sets the time spent at each step in the source sweep in seconds.
                    Must be a multiple of 200 microseconds (0.0002).
                direction (Direction):
                    The direction of the sweep.
                    UP begins the sweep at the start and ends at the stop value.
                    DOWN begins the sweep at the stop value and ends at the start value.
                round_trip (bool):
                    The round trip state of the sweep.
                    When True, the sweep will begin and end at the same value.
                spacing (SweepSpacing):
                    The spacing of the sweep.
            """

            self.sweep_type = sweep_type
            self.spacing = spacing
            self.start = start
            self.stop = stop
            self.points = points
            self.dwell = dwell
            self.direction = direction
            self.round_trip = round_trip



"""Implements functionality unique to the Lake Shore Model 224 temperature monitor"""

from enum import IntEnum
import serial
from .generic_instrument import GenericInstrument, InstrumentException, RegisterBase
from .temperature_controllers import StandardEventRegister


class Model224AlarmParameters:
    """Class used to disable or configure an alarm in conjunction with the set/get_alarm_parameters() method"""

    def __init__(self, high_value, low_value, deadband, latch_enable, audible=None, visible=None):
        """Constructor for Model224AlarmParameters class

            Args:
                high_value (float):
                    Sets the value the source is checked against to activate the
                    high alarm
                low_value (float):
                    Sets the value the source is checked against to activate low alarm.
                deadband (float):
                    Sets the value that the source must change outside of an alarm
                    condition to deactivate an unlatched alarm.
                latch_enable (bool):
                    Specifies a latched alarm (False = off, True = on)
                audible (bool):
                    Specifies if the internal speaker will beep when an alarm condition
                    occurs (False = off, True = on)
                    Optional parameter.
                visible (bool):
                    Specifies if the Alarm LED on the instrument front panel will blink
                    when an alarm condition occurs (False = off, True = on)
                    Optional parameter.
        """
        self.high_value = high_value
        self.low_value = low_value
        self.deadband = deadband
        self.latch_enable = latch_enable
        self.audible = audible
        self.visible = visible


class Model224InputSensorType(IntEnum):
    """Enumeration for the type of sensor being used for a given input."""
    INPUT_DISABLED = 0
    DIODE = 1
    PLATINUM_RTD = 2
    NTC_RTD = 3


class Model224InputSensorUnits(IntEnum):
    """Enumeration for the preferred units of an input sensor."""
    KELVIN = 1
    CELSIUS = 2
    SENSOR = 3


class Model224DiodeSensorRange(IntEnum):
    """Enumeration for the voltage range of a diode sensor."""
    RANGE_2_POINT_5_VOLTS = 0
    RANGE_10_VOLTS = 1


class Model224PlatinumRTDSensorResistanceRange(IntEnum):
    """Enumeration of the resistance range of a platinum RTD input sensor."""
    TEN_OHMS = 0
    THIRTY_OHMS = 1
    ONE_HUNDRED_OHMS = 2
    THREE_HUNDRED_OHMS = 3
    ONE_KILOHM = 4
    THREE_KILOHMS = 5
    TEN_KILOHMS = 6


class Model224NTCRTDSensorResistanceRange(IntEnum):
    """Enumeration of the resistance range of a NTC RTD input sensor."""
    TEN_OHMS = 0
    THIRTY_OHMS = 1
    ONE_HUNDRED_OHMS = 2
    THREE_HUNDRED_OHMS = 3
    ONE_KILOHM = 4
    THREE_KILOHMS = 5
    TEN_KILOHMS = 6
    THIRTY_KILOHMS = 7
    ONE_HUNDRED_KILOHMS = 8


class Model224InputSensorSettings:
    """Class representing the parameters of a sensor in one of the instrument's inputs."""

    def __init__(self,
                 sensor_type,
                 preferred_units,
                 sensor_range=None,
                 autorange_enabled=False,
                 compensation=False):
        """Constructor for the Model224InputSensorSettings class.

            Args:
                sensor_type (Model224InputSensorType or int):
                    * Specifies what type of sensor is being used at the input.
                preferred_units (Model224InputSensorUnits or int):
                    * Specifies the preferred units used for sensor readings and alarm setpoints when displayed.
                sensor_range (IntEnum):
                    * Specifies the range of the sensor.
                    * Optional if auto range is enabled
                autorange_enabled (bool):
                    * Defines if autorange is enabled.
                    * Not applicable for diode sensors
                    * Defaults to false
                compensation (bool):
                    * Defines if thermal input compensation is on or off.
                    * Not applicable for diode sensors
                    * Defaults to false
        """
        self.sensor_type = sensor_type
        self.sensor_range = sensor_range
        self.preferred_units = preferred_units
        self.autorange_enabled = autorange_enabled
        self.compensation = compensation


class Model224InterfaceMode(IntEnum):
    """Enumeration for the mode of the remote interface"""
    LOCAL = 0
    REMOTE = 1
    REMOTE_LOCAL_LOCK = 2


class Model224RemoteInterface(IntEnum):
    """Enumeration for the remote interface being used to communicate with the instrument."""
    USB = 0
    ETHERNET = 1
    IEEE_488 = 2


class Model224DisplayFieldUnits(IntEnum):
    """Enumerated type defining how units are enumerated for settings and using Display Fields."""
    KELVIN = 1
    CELSIUS = 2
    SENSOR = 3
    MINIMUM_DATA = 4
    MAXIMUM_DATA = 5


class Model224InputChannel(IntEnum):
    """Enumerated type defining which input channels correspond to ints for setting and using Display Fields."""
    NO_INPUT = 0
    INPUT_A = 1
    INPUT_B = 2
    INPUT_C = 3
    INPUT_D1 = 4
    INPUT_D2 = 5
    INPUT_D3 = 6
    INPUT_D4 = 7
    INPUT_D5 = 8
    INPUT_C2 = 9
    INPUT_C3 = 10
    INPUT_C4 = 11
    INPUT_C5 = 12


class Model224DisplayMode(IntEnum):
    """Enumeration defining what input or information is shown on the front panel display."""
    INPUT_A = 0
    INPUT_B = 1
    INPUT_C = 2
    INPUT_D1 = 3
    CUSTOM = 4
    ALL_INPUTS = 5
    INPUT_D2 = 6
    INPUT_D3 = 7
    INPUT_D4 = 8
    INPUT_D5 = 9
    INPUT_C2 = 10
    INPUT_C3 = 11
    INPUT_C4 = 12
    INPUT_C5 = 13


class Model224NumberOfFields(IntEnum):
    """Enumerated type specifying the number of display fields to configure in the Custom display mode."""
    LARGE_4 = 0
    LARGE_8 = 1
    LARGE_4_SMALL_8 = 2
    SMALL_16 = 3


class Model224RelayControlAlarm(IntEnum):
    """Enumeration of the setting determining which alarm(s) cause a relay to activate in alarm mode."""
    LOW_ALARM = 0
    HIGH_ALARM = 1
    BOTH_ALARMS = 2


class Model224RelayControlMode(IntEnum):
    """Enumeration of the configured mode of a relay."""
    RELAY_OFF = 0
    RELAY_ON = 1
    ALARMS = 2


class Model224CurveFormat(IntEnum):
    """Enumerations specify formats for temperature sensor curves"""
    MILLIVOLT_PER_KELVIN = 1
    VOLTS_PER_KELVIN = 2
    OHMS_PER_KELVIN = 3
    LOG_OHMS_PER_KELVIN = 4


class Model224CurveTemperatureCoefficients(IntEnum):
    """Enumerations specify positive/negative temperature sensor curve coefficients"""
    NEGATIVE = 1
    POSITIVE = 2


class Model224DiodeExcitationCurrent(IntEnum):
    """Enum type representing the different excitation currents available for a diode sensor."""
    TEN_MICRO_AMPS = 0
    ONE_MILLI_AMP = 1


class Model224SoftCalSensorTypes(IntEnum):
    """Enum type representing the standard curves used to generate a SoftCal curve. The 3 standard curves each
    represent a different type of sensor that can be calibrated with a SoftCal curve."""
    DT_400 = 1
    PT_100 = 6
    PT_1000 = 7


class Model224CurveHeader:
    """A class that configures the user curve header and corresponding parameters"""

    def __init__(self, curve_name, serial_number, curve_data_format, temperature_limit, coefficient):
        """Constructor for Model224CurveHeader class

            Args:
                curve_name (str):
                    Specifies curve name (limit of 15 characters)

                serial_number (str):
                    Specifies curve serial number (limit of 10 characters)

                curve_data_format (Model224CurveFormat):
                    Specifies the curve data format

                temperature_limit (float):
                    * Specifies the curve temperature limit in Kelvin

                coefficient (Model224CurveTemperatureCoefficients):
                    * Specifies the curve temperature coefficient

        """

        self.curve_name = curve_name
        self.serial_number = serial_number
        self.curve_data_format = curve_data_format
        self.temperature_limit = temperature_limit
        self.coefficient = coefficient


Model224StandardEventRegister = StandardEventRegister


class Model224ServiceRequestRegister(RegisterBase):
    """Class object representing the Service Request Enable register."""
    bit_names = [
        "",
        "",
        "",
        "",
        "message_available",
        "event_summary",
        ""
        "operation_summary"
    ]

    def __init__(self,
                 message_available,
                 event_summary,
                 operation_summary):
        self.message_available = message_available
        self.event_summary = event_summary
        self.operation_summary = operation_summary


class Model224StatusByteRegister(RegisterBase):
    """Class object representing the status byte register."""
    bit_names = [
        "",
        "",
        "",
        "",
        "message_available",
        "event_summary",
        "master_summary_status"
        "operation_summary"
    ]

    def __init__(self,
                 message_available,
                 event_summary,
                 master_summary_status,
                 operation_summary):
        self.message_available = message_available
        self.event_summary = event_summary
        self.master_summary_status = master_summary_status
        self.operation_summary = operation_summary


class Model224ReadingStatusRegister(RegisterBase):
    """Class object representing the reading status of an input. While not a literal register, the return of an int
    representation of multiple booleans makes it conveninet to represent this functionality as a register."""
    bit_names = [
        "invalid_reading",
        "",
        "",
        "",
        "temperature_under_range",
        "temperature_over_range",
        "sensor_units_zero",
        "sensor_units_over_range"
    ]

    def __init__(self,
                 invalid_reading,
                 temperature_under_range,
                 temperature_over_range,
                 sensor_units_zero,
                 sensor_units_over_range):
        self.invalid_reading = invalid_reading
        self.temperature_under_range = temperature_under_range
        self.temperature_over_range = temperature_over_range
        self.sensor_units_zero = sensor_units_zero
        self.sensor_units_over_range = sensor_units_over_range


class Model224(GenericInstrument):
    """A class object representing the Lake Shore Model 224 temperature monitor"""

    vid_pid = [(0x1FB9, 0x0204)]

    def __init__(self,
                 serial_number=None,
                 com_port=None,
                 baud_rate=57600,
                 data_bits=7,
                 stop_bits=1,
                 parity=serial.PARITY_ODD,
                 flow_control=False,
                 handshaking=False,
                 timeout=2.0,
                 ip_address=None,
                 tcp_port=7777,
                 **kwargs):

        # Call the parent init, then fill in values specific to the 224
        GenericInstrument.__init__(self, serial_number, com_port, baud_rate, data_bits, stop_bits, parity, flow_control,
                                   handshaking, timeout, ip_address, tcp_port, **kwargs)

    @staticmethod
    def _error_check(error_code):
        event_register = Model224StandardEventRegister.from_integer(error_code)
        if event_register.query_error:
            raise InstrumentException('Query Error')
        if event_register.command_error:
            raise InstrumentException('Command Error: Invalid Command or Query')
        if event_register.execution_error:
            raise InstrumentException('Execution Error: Instrument not able to execute command or query.')

    def command(self, *commands, check_errors=True):
        """Send a SCPI command or multiple commands to the instrument

            Args:
                commands (str):
                    * A serial command

            Kwargs:
                check_errors (bool):
                    * Chooses whether to check for and raise errors after sending a command. True by default.

        """

        # Group all commands and queries a single string with SCPI delimiters.
        command_string = ";:".join(commands)

        if check_errors:
            self.query(command_string)
        else:
            command_string += ";*OPC?"
            self.query(command_string, check_errors=False)

    def query(self, *queries, check_errors=True):
        """Send a query to the instrument and return the response

            Args:
                queries (str):
                    * A serial query ending in a question mark

            Return:
               * The instrument query response as a string.

        """

        # Group all commands and queries a single string with SCPI delimiters.
        query_string = ";:".join(queries)

        # Append the query with an additional error buffer query.
        if check_errors:
            query_string += ";*ESR?"

        response = GenericInstrument.query(self, query_string)

        if check_errors:
            response_list = response.split(';')
            error_code = response_list.pop()
            self._error_check(error_code)
            response = ';'.join(response_list)

        return response

    def get_standard_event_enable_mask(self):
        """Returns the names of the standard event enable register bits and their values.
        These values determine which bits propagate to the standard event register"""

        response = self.query("*ESE?")
        status_register = Model224StandardEventRegister.from_integer(response)
        return status_register

    def set_standard_event_enable_mask(self, register_mask):
        """Configures values of the standard event enable register bits.
        These values determine which bits propagate to the standard event register

            Args:
                register_mask (Model224StandardEventRegister):
                    An StandardEventRegister class object with all bits set to a value

        """

        integer_representation = register_mask.to_integer()
        self.command("*ESE " + str(integer_representation))

    def clear_interface_command(self):
        """Clears the bits in the Status Byte Register, Standard Event Status Register, and Operation Event Register,
        and terminates all pending operations. Clears the interface, but not the controller."""

        self.command("*CLS")

    def reset_instrument(self):
        """Sets controller parameters to power-up settings"""

        self.command("*RST")

    def set_service_request(self, register_mask):
        """Manually enable/disable the mask of the corresponding status flag bit in the status
        byte register

            Args:
                register_mask (Model224ServiceRequestRegister):
                    A Model224ServiceRequestRegister class object with all bits configured
        """

        integer_representation = register_mask.to_integer()
        self.command("*SRE " + str(integer_representation))

    def get_service_request(self):
        """Returns the status byte register bits and their values as a class instance"""

        response = self.query("*SRE?")
        status_register = Model224ServiceRequestRegister.from_integer(response)
        return status_register

    def get_status_byte(self):
        """Returns the status flag bits as a class instance without resetting the register"""

        response = self.query("*STB?")
        status_flag = Model224StatusByteRegister.from_integer(response)
        return status_flag

    def get_self_test(self):
        """Instrument self test result completed at power up

            Return:
                test_errors (bool):
                    * True = errors found
                    * False = no errors found
        """

        test_errors = bool(int(self.query("*TST?")))
        return test_errors

    def set_wait_to_continue(self):
        """Causes the IEEE-488 interface to hold off until all pending operations have been completed.
        This has the same function as the set_operation_complete() method, except that it does not set the
        Operation Complete event bit in the Event Status Register"""

        self.command("*WAI")

    def set_to_factory_defaults(self):
        """Sets all the settings and configurations to their factory default values."""
        self.command("DFLT 99")

    def get_reading_status(self, input_channel):
        """Returns the reading status of any input status flags that may be set.

            Args:
                input_channel (str):
                    * The input to check for reading status flags.
                    * Options are:
                        * A
                        * B
                        * C(1 - 5)
                        * D(1 - 5)

            Returns:
                (dict):
                    {invalid_reading: bool, temperature_under_range: bool, temperature_over_range: bool,
                    sensor_units_zero: bool, sensor_units_over_range: bool}
        """
        flag_code = int(self.query(f"RDGST? {input_channel}"))
        reading_status = Model224ReadingStatusRegister.from_integer(flag_code)
        return reading_status

    def get_kelvin_reading(self, input_channel):
        """Returns the temperature value in kelvin of either channel

            Args:
                input_channel:
                    * Selects the channel to retrieve measurement.
                    * Options are:
                        * A
                        * B
                        * C(1 - 5)
                        * D(1 - 5)

            Returns:
                (float):
                    The reading of the sensor in kelvin


        """

        return float(self.query(f"KRDG? {input_channel}"))

    def get_sensor_reading(self, input_channel):
        """Returns the sensor reading in the sensor's units.

            Args:
                input_channel:
                    * Selects the channel to retrieve measurement.
                    * Options are:
                        * A
                        * B
                        * C(1 - 5)
                        * D(1 - 5)

            Returns:
                reading (float):
                    * The raw sensor reading in the units of the connected sensor

        """

        return float(self.query(f"SRDG? {input_channel}"))

    def get_celsius_reading(self, input_channel):
        """Returns the given input's temperature reading in degrees Celsius.

            Args:
                input_channel (str)
                    Selects input to retrieve measurement from.

            Returns:
                (float):
                    Temperature readings in degrees Celsius
        """
        return float(self.query(f"CRDG? {input_channel}"))

    def get_all_inputs_celsius_reading(self):
        """Returns the temperature reading in degrees Celsius of all the inputs.

            Returns:
                (dict):
                    {input_a_reading: float, input_b_reading: float, input_c1_reading: float, input_c2_reading: float,
                    input_c3_reading: float, input_c4_reading: float, input_c5_reading: float, input_d1_reading: float,
                    input_d2_reading: float, input_d3_reading: float, input_d4_reading: float, input_d5_reading: float}
        """
        reading = self.query("CRDG? 0")
        separated_readings = reading.split(",")
        return {'input_a_reading': float(separated_readings[0]),
                'input_b_reading': float(separated_readings[1]),
                'input_c1_reading': float(separated_readings[2]),
                'input_c2_reading': float(separated_readings[3]),
                'input_c3_reading': float(separated_readings[4]),
                'input_c4_reading': float(separated_readings[5]),
                'input_c5_reading': float(separated_readings[6]),
                'input_d1_reading': float(separated_readings[7]),
                'input_d2_reading': float(separated_readings[8]),
                'input_d3_reading': float(separated_readings[9]),
                'input_d4_reading': float(separated_readings[10]),
                'input_d5_reading': float(separated_readings[11])}

    def set_input_diode_excitation_current(self, input_channel, diode_current):
        """Sets the excitation current of a diode sensor. Input must be configured for a diode sensor for command
        to work. Current defaults to 10uA.

            Args:
                input_channel (str):
                    * The input to configure the diode excitation current for.

                diode_current (Model224DiodeExcitationCurrent):
                    * The excitation current for the diode sensor.

        """
        self.command(f"DIOCUR {input_channel},{diode_current}")

    def get_input_diode_excitation_current(self, input_channel):
        """Returns the diode excitation current for the given diode sensor.

            Args:
                input_channel (str):
                    The diode sensor input to query the current of.

            Returns:
                diode_current (Model224DiodeExcitationCurrent):
                    A member of the Model224DiodeExcitationCurrent enum class.

        """
        diode_current_int = int(self.query(f"DIOCUR? {input_channel}"))
        return Model224DiodeExcitationCurrent(diode_current_int)

    def set_sensor_name(self, channel, sensor_name):
        """Sets a given name to a sensor on the specified channel

            Args:
                channel (str):
                    * Specifies which the sensor to name is on.
                    * Options are:
                        * A
                        * B
                        * C(1 - 5)
                        * D(1 - 5)

                sensor_name(str):
                    * Name user wants to give to the sensor on the specified channel

        """

        self.command(f"INNAME {channel},\"{sensor_name}\"")

    def get_sensor_name(self, channel):
        """Returns the name of the sensor on the specified channel

            Args:
                channel (str):
                    * Specifies which input sensor to retrieve name of.
                    * Options are:
                        * A
                        * B
                        * C(1 - 5)
                        * D(1 - 5)

            Returns:
                name (str)
                    * Name associated with the sensor

        """

        return self.query(f"INNAME? {channel}")

    def set_display_contrast(self, contrast_level):
        """Sets the contrast level for the front panel display

            Args:
                contrast_level (int):
                    * Display contrast for the front panel LCD screen
                    * Options are:
                        * 1 - 32

        """

        self.command(f"BRIGT {contrast_level}")

    def get_display_contrast(self):
        """Returns the contrast level of front panel display

            Return:
                (int):
                    * Contrast level of the front panel LCD screen

        """

        return int(self.query("BRIGT?"))

    def set_ieee_488(self, address):
        """Specifies the IEEE address

            Args:
                address (int):
                    * 1-30 (0 and 31 reserved)
        """
        self.command(f"IEEE {address}")

    def get_ieee_488(self):
        """Returns the IEEE address set

            Return:
                address (int):
                    * 1-30 (0 and 31 reserved)

        """
        return int(self.query("IEEE?"))

    def set_led_state(self, state):
        """Sets the front panel LEDs to on or off.

            Args:
                state (bool)
                    * Sets the LEDs to functional or nonfunctional. Options are:
                    * False for off or True for on

        """
        self.command(f"LEDS {str(int(state))}")

    def get_led_state(self):
        """Returns whether or not front panel LEDs are enabled.

            Returns:
                state (bool)
                    * Specifies whether front panel LEDs are functional. Returns:
                    * False if disabled, True enabled.

        """
        return bool(int(self.query("LEDS?")))

    def set_keypad_lock(self, state, code):
        """Locks or unlocks front panel keypad (except for alarms and disabling heaters).

            Args:
                state (bool)
                    * Sets the keypad to locked or unlocked. Options are:
                    * False for unlocked or True for locked

                code (int)
                    * Specifies 3 digit lock-out code. Options are:
                    * 000 - 999

        """
        self.command(f"LOCK {str(int(state))},{str(code)}")

    def get_keypad_lock(self):
        """Returns the state of the keypad lock and the lock-out code.

            Return:
                (dict):
                    * [state: bool, code: int]

        """
        output_string = self.query("LOCK?")
        separated_response = output_string.split(",")
        return {'state': bool(int(separated_response[0])),
                'code': int(separated_response[1])}

    def get_min_max_data(self, input_channel):
        """Returns the minimum and maximum data from an input

            Args:
                input_channel (str):
                    Specifies which input to query
            Return:
                min_max_data (dict):
                    * [minimum: float, maximum: float]

        """
        min_max_data = self.query("MDAT? " + str(input_channel)).split(",")
        min_max_dictionary = {"minimum": float(min_max_data[0]),
                              "maximum": float(min_max_data[1])}
        return min_max_dictionary

    def reset_min_max_data(self):
        """Resets the minimum and maximum input data"""
        self.command("MNMXRST")

    def set_input_curve(self, input_channel, curve_number):
        """Specifies the curve an input uses for temperature conversion

            Args:
                input_channel (str):
                    Specifies which input to configure

                curve_number (int):
                    * 0 = none, 1-20 = standard curves, 21-59 = user curves

        """
        self.command(f"INCRV {str(input_channel)},{str(curve_number)}")
        # Check that the user mapped an input to a curve (not just set the input to no curve)
        if curve_number != 0:
            # Query the curve mapped to input_channel, if the query returns zero,
            # an invalid curve was selected for the specified input
            set_curve = self.get_input_curve(input_channel)
            if set_curve == 0:
                raise InstrumentException("The specified curve type does not match the configured input type")

    def get_input_curve(self, input_channel):
        """Returns the curve number being used for a given input

            Args:
                input_channel (str):
                    Specifies which input to query

            Return:
                curve_number (int):
                    * 0-59

        """
        return int(self.query(f"INCRV? {str(input_channel)}"))

    def set_website_login(self, username, password):
        """Sets the username and password to connect instrument to website.

            Args:
                username (str)
                    * Username to set for login.
                    * Must be less than or equal to 15 characters.
                    * Method automatically puts quotation marks around string, so they are not needed in the
                      string literal passed into the method.

                password (str)
                    * password to set for login.
                    * Must be less than or equal to 15 characters.
                    * Method automatically puts quotation marks around string, so they are not needed in the
                      string literal passed into the method.

        """
        self.command("WEBLOG \"" + username + "\",\"" + password + "\"")

    def get_website_login(self):
        """Returns the set username and password for web login for the instrument.

            Returns:
                (dict):
                    {username: str, password: str}
        """
        username_password = self.query("WEBLOG?")
        separated_string = username_password.split(",")
        # Remove padded whitespace and quotations in the returned username and password
        username = separated_string[0].strip(' "')
        password = separated_string[1].strip(' "')
        return {"username": username,
                "password": password}

    def set_alarm_parameters(self, input_channel, alarm_enable, alarm_settings=None):
        """Configures the alarm parameters for an input

            Args:
                input_channel (str):
                    Specifies which input to configure
                alarm_enable (bool):
                    Specifies whether to turn on the alarm for the input, or turn the alarm off.
                alarm_settings (Model224AlarmParameters):
                    See Model224AlarmParameters class. Optional if alarm_enable is set to False

        """

        if alarm_enable:
            if alarm_settings.audible is None:
                audible = 0
            else:
                audible = int(alarm_settings.audible)
            if alarm_settings.visible is None:
                visible = 0
            else:
                visible = int(alarm_settings.visible)
            self.command(f"ALARM {input_channel.upper()},{int(alarm_enable)},{alarm_settings.high_value}," +
                        f"{alarm_settings.low_value},{alarm_settings.deadband},{int(alarm_settings.latch_enable)}," +
                        f"{audible},{visible}")
        else:
            self.command("ALARM " + str(input_channel.upper()) + "," + str(int(alarm_enable)) + ",0,0,0,0,0,0")

    def get_alarm_parameters(self, input_channel):
        """Returns the present state of all alarm parameters

            Args:
                input_channel (str):
                    Specifies which input to configure

            Return:
                (dict):
                    {alarm_enable: bool, alarm_settings: Model224AlarmParameters}

        """
        parameters = self.query("ALARM? " + str(input_channel)).split(",")
        alarm_enable = bool(int(parameters[0]))
        alarm_settings = Model224AlarmParameters(float(parameters[1]), float(parameters[2]),
                                                 float(parameters[3]), bool(int((parameters[4]))),
                                                 audible=bool(int(parameters[5])), visible=bool(int(parameters[6])))
        return {'alarm_enable': alarm_enable,
                'alarm_settings': alarm_settings}

    def get_alarm_status(self, input_channel):
        """Returns the high state and low state of the alarm for the specified channel

            Args:
                input_channel (str)
                    * Specifies which input channel to read from.
                    * Options are:
                        * A
                        * B
                        * C(1 - 5)
                        * D(1 - 5)

            Returns:
                (dict):
                    {high_state: bool, low_state: bool}
                        * high_state (bool)
                            * True if high state is on, False if high state is off

                        * low_state (bool)
                            * True if low state is on, False if low state is off

        """
        response = self.query("ALARMST? " + str(input_channel))
        separated_response = response.split(",")
        return {"high_state": bool(int(separated_response[0])),
                "low_state": bool(int(separated_response[1]))}

    def reset_alarm_status(self):
        """Clears the high and low status of all alarms."""
        self.command("ALMRST")

    def set_curve_header(self, curve_number, curve_header):
        """Configures the user curve header

            Args:
                curve_number (int):
                    * Specifies which curve to configure.
                    * Options are:
                        * 21 - 59

                curve_header (Model224CurveHeader):
                    * A Model224CurveHeader class object containing the desired curve information

        """
        command_string = (f"CRVHDR {curve_number},{curve_header.curve_name},{curve_header.serial_number}," +
                        f"{curve_header.curve_data_format},{curve_header.temperature_limit},{curve_header.coefficient}")

        self.command(command_string)

    def get_curve_header(self, curve):
        """Returns parameters set on a particular user curve header

            Args:
                curve (int):
                    * Specifies a curve to retrieve
                    * Options are:
                        * 21 - 59

            Returns:
                header (Model224CurveHeader):
                    * A Model224CurveHeader class object containing the desired curve information

        """
        response = self.query(f"CRVHDR? {curve}")
        curve_header = response.split(",")
        header = Model224CurveHeader(str(curve_header[0]),
                                     str(curve_header[1]),
                                     Model224CurveFormat(int(curve_header[2])),
                                     float(curve_header[3]),
                                     Model224CurveTemperatureCoefficients(int(curve_header[4])))

        return header

    def set_curve_data_point(self, curve, index, sensor_units, temperature):
        """Configures a user curve point

            Args:
                curve (int or str):
                    * Specifies which curve to configure

                index (int):
                    * Specifies the points index in the curve

                sensor_units (float):
                    * Specifies sensor units for this point to 6 digits

                temperature (float):
                    * Specifies the corresponding temperature in Kelvin for this point to 6 digits

        """
        self.command(f"CRVPT {curve},{index},{sensor_units},{temperature}")

    def get_curve_data_point(self, curve, index):
        """Returns a standard or user curve data point

            Args:
                curve (int):
                    * Specifies which curve to query

                index (int):
                    * Specifies the points index in the curve

            Return:
                curve_point (tuple)
                    * (sensor_units: float, temp_value: float))

        """
        curve_point = self.query(f"CRVPT? {curve},{index}").split(",")
        return float(curve_point[0]), float(curve_point[1])

    def delete_curve(self, curve):
        """Deletes the user curve

            Args:
                curve (int):
                    * Specifies a user curve to delete

        """
        self.command(f"CRVDEL {curve}")

    def generate_and_apply_soft_cal_curve(self, source_curve, curve_number, serial_number, calibration_point_1,
                                          calibration_point_2=(0, 0), calibration_point_3=(0, 0)):
        """Creates a SoftCal curve from 1-3 temperature/sensor points and a standard curve. Inputs generated curve
        into the given curve number.

            Args:
                source_curve (Model224SoftCalSensorTypes):
                    * The standard curve to use to generate the SoftCal curve from along with calibration points.

                curve_number (int):
                    * The curve number to save the generated curve to.
                    * Options are:
                        * 21 - 59

                serial_number (str):
                    * Serial number of the user curve.
                    * Maximum of 10 characters

                calibration_point_1 (tuple):
                    * Tuple of two floats in the form (temperature_value, sensor_value)

                calibration_point_2 (tuple):
                    * Tuple of two floats in the form (temperature_value, sensor_value)
                    * Optional parameter

                calibration_point_3 (tuple):
                    * Tuple of two floats in the form (temperature_value, sensor_value)
                    * Optional parameter

        """
        command_string = (f"SCAL {source_curve},{curve_number},{serial_number}," +
                            f"{calibration_point_1[0]},{calibration_point_1[1]}," +
                            f"{calibration_point_2[0]},{calibration_point_2[1]}," +
                            f"{calibration_point_3[0]},{calibration_point_3[1]}")
        self.command(command_string)

    def get_curve(self, curve):
        """Returns a list of all the data points in a particular curve

            Args:
                curve (int):
                    * Specifies which curve to set

            Return:
                data_points (list):
                    * A list containing every point in the curve represented as a tuple
                        * (sensor_units: float, temp_value: float)

        """

        data_points = []
        true_point_index = 0
        for i in range(0, 200):
            point = self.get_curve_data_point(curve, i + 1)
            data_points.append(point)
            if point[0] != 0 or point[1] != 0:
                true_point_index = i

        # Remove all extraneous points
        data_points = data_points[:true_point_index + 1]

        return data_points

    def set_curve(self, curve, data_points):
        """Method to define a user curve using a list of data points

            Args:
                curve (int):
                    * Specifies which curve to set

                data_points (list):
                    * A list containing every point in the curve represented as a tuple
                        * (sensor_units: float, temp_value: float)

        """

        self.delete_curve(curve)

        for index, point in data_points:
            self.set_curve_data_point(curve, index + 1, point[0], point[1])

    def get_relay_status(self, relay_channel):
        """Returns whether the specified relay is On or Off.

            Args:
                relay_channel (int)
                    * The relay channel to query.
                    * Options are:
                        * 1 or 2

            Returns:
                (bool):
                    * True if relay is on, False if relay is off.
        """
        return bool(int(self.query(f"RELAYST? {str(relay_channel)}")))

    def set_filter(self, input_channel, filter_enabled, number_of_points=8, filter_reset_threshold=10):
        """Enables or disables a filter for the readings of the specified input channel. Filter is a running
        average that smooths input readings exponentially.

            Args:
                input_channel (str):
                    * The input to set or disable a filter for.
                    * Options are:
                        * A
                        * B
                        * C(1 - 5)
                        * D(1 - 5)

                filter_enabled (bool):
                    * Enables or disables a filter for the input channel.
                    * True for enabled, False for disabled.

                number_of_points (int):
                    * Specifies the number of points used for the filter.
                    * Inputting a larger number of points will slow down the instrument's response to changes in
                        temperature.
                    * Options are:
                        * 2 - 64
                    * Optional if disabling the filter function.

                filter_reset_threshold (int):
                    * Specifies the limit for restarting the filter, represented by a percent of the full scale reading.
                        If raw reading differs from filtered value by more than this threshold, filter averaging resets.
                    * Options are:
                        * 1% - 10%
                    * Optional if disabling the filter function.

        """
        self.command("FILTER " + str(input_channel) + "," + str(int(filter_enabled)) + "," + str(number_of_points) +
                     "," + str(filter_reset_threshold))

    def get_filter(self, input_channel):
        """Retrieves information about the filter set on the specified input channel.

            Args:
                input_channel (str):
                    * The input to query for filter information.
                    * Options are:
                        * A
                        * B
                        * C(1 - 5)
                        * D(1 - 5)

            Returns:
                (dict):
                    {filter_enabled: bool, number_of_points: int, filter_reset_threshold: int}

        """
        filter_information = self.query(f"FILTER? {str(input_channel)}")
        separated_information = filter_information.split(",")
        return {'filter_enabled': bool(int(separated_information[0])),
                'number_of_points': int(separated_information[1]),
                'filter_reset_threshold': int(separated_information[2])}

    def configure_input(self, input_channel, settings):
        """Configures a sensor for measurement input readings.

            Args:
                input_channel (str):
                    The input to configure the input for. Options are:
                        * A
                        * B
                        * C(1 - 5)
                        * D(1 - 5)

                settings (Model224InputSensorSettings):
                    Object of the Model224InputSensorSettings containing information for sensor setup.

        """
        command_string = (f"INTYPE {input_channel},{settings.sensor_type},{int(settings.autorange_enabled)}," +
                        f"{settings.sensor_range},{int(settings.compensation)},{settings.preferred_units}")
        self.command(command_string)

    def disable_input(self, input_channel):
        """Disables the selected input channel.

            Args:
                input_channel (str):
                    The input to disable. Options are:
                        * A
                        * B
                        * C (1 - 5)
                        * D (1 - 5)

        """
        # Fill all parameters with 0 to disable
        self.command(f"INTYPE {str(input_channel)},0,0,0,0,0")

    def get_input_configuration(self, input_channel):
        """Returns the configuration settings of the sensor at the specified input channel.

            Args:
                input_channel (str)
                    The input to query. Options are:
                        * A
                        * B
                        * C(1 - 5)
                        * D(1 - 5)

            Returns:
                (Model224InputSensorSettings):
                    Object of type Model224InputSensorSettings containing information about the sensor at the given
                    input_channel

        """
        settings_string = self.query(f"INTYPE? {str(input_channel)}")
        separated_settings = settings_string.split(",")
        # Convert sensor_range depending on sensor type
        sensor_type = int(separated_settings[0])
        if sensor_type == 1:
            sensor_range = Model224DiodeSensorRange(int(separated_settings[2]))
        elif sensor_type == 2:
            sensor_range = Model224PlatinumRTDSensorResistanceRange(int(separated_settings[2]))
        elif sensor_type == 3:
            sensor_range = Model224NTCRTDSensorResistanceRange(int(separated_settings[2]))
        else:
            # Sensor is disabled
            sensor_range = None
        # Create object
        return Model224InputSensorSettings(Model224InputSensorType(sensor_type),
                                           Model224InputSensorUnits(int(separated_settings[4])),
                                           sensor_range,
                                           bool(int(separated_settings[1])),
                                           bool(int(separated_settings[3])))

    def select_remote_interface(self, remote_interface):
        """Selects the remote interface to use for communications.

            Args:
                remote_interface (Model224RemoteInterface):
                    Object of enum type Model224RemoteInterface, representing the type of interface used for
                    communications

        """
        self.command(f"INTSEL {remote_interface}")

    def get_remote_interface(self):
        """Returns the remote interface being used for communications.

            Returns:
                (Model224RemoteInterface):
                    Object of enum type Model224RemoteInterface representing the interface being used for
                    communications

        """
        interface_number = int(self.query("INTSEL?"))
        return Model224RemoteInterface(interface_number)

    def select_interface_mode(self, interface_mode):
        """Selects the mode for the remote interface being used.

            Args:
                interface_mode (Model224InterfaceMode):
                    Object of enum type Model224InterfaceMode representing the desired communication mode.

        """
        self.command(f"MODE {interface_mode}")

    def get_interface_mode(self):
        """Returns the mode of the remote interface.

            Returns:
                (Model224InterfaceMode):
                    Object of enum type Model224InterfaceMode representing the communication mode.

        """
        mode_number = int(self.query("MODE?"))
        return Model224InterfaceMode(mode_number)

    def set_display_field_settings(self, field, input_channel, display_units):
        """Configures a display field in custom display mode.

            Args:
                field (int):
                    * Specifies which display field to configure.
                    * Options are:
                        * 1 - 8

                input_channel (Model224InputChannel)
                    * Defines which input to display.

                display_units (Model224DisplayFieldUnits)
                    * Defines which units to display reading in.

        """
        self.command(f"DISPFLD {field},{input_channel},{display_units}")

    def get_display_field_settings(self, field):
        """Returns the settings of a single display field in custom display mode.

            Args:
                field (int):
                    * Specifies the display field to query.
                    * Options are:
                        * 1 - 8

            Returns:
                (dict):
                    {input_channel: Model224InputChannel, display_units: Model224DisplayFieldUnits}

        """
        display_field_settings = self.query("DISPFLD? " + str(field))
        separated_settings = display_field_settings.split(",")
        return {'input_channel': Model224InputChannel(int(separated_settings[0])),
                'display_units': Model224DisplayFieldUnits(int(separated_settings[1]))}

    def configure_display(self, display_mode, number_of_fields=0):
        """Configures the display of the instrument.

            Args:
                display_mode (Model224DisplayMode):
                    * Defines what mode to set the display in.
                    * Mode either defines which input to display, or sets up a custom display using display fields.

                number_of_fields (Model224NumberOfFields):
                    * Defines the number of display locations to display.
                    * Only valid if mode is set to CUSTOM

        """
        self.command(f"DISPLAY {display_mode},{number_of_fields}")

    def get_display_configuration(self):
        """Returns the mode of the display. If display mode is Custom, this method also returns the number of
        display fields in the custom display.

            Returns:
                (dict):
                    {display_mode: Model224DisplayMode, number_of_fields: Model224NumberOfFields}

        """
        display_settings_string = self.query("DISPLAY?")
        separated_settings = display_settings_string.split(",")
        display_mode = int(separated_settings[0])
        if display_mode != 4:
            return_dictionary = {'display_mode': Model224DisplayMode(display_mode),
                                 'number_of_fields': None}
        else:
            number_of_fields = int(separated_settings[1])
            return_dictionary = {'display_mode': Model224DisplayMode(display_mode),
                                 'number_of_fields': Model224NumberOfFields(number_of_fields)}
        return return_dictionary

    def turn_relay_on(self, relay_number):
        """Turns the specified relay on.

            Args:
                relay_number (int):
                    * The relay to turn on.
                    * Options are:
                        * 1 or 2

        """
        self.command(f"RELAY {relay_number},1,0,0")

    def turn_relay_off(self, relay_number):
        """Turns the specified relay off.

            Args:
                relay_number (int):
                    * The relay to turn off.
                    * Options are:
                        * 1 or 2

        """
        self.command(f"RELAY {relay_number},0,0,0")

    def set_relay_alarms(self, relay_number, activating_input_channel, alarm_relay_trigger_type):
        """Sets a relay to turn on and off automatically based on the state of the alarm of the specified input channel.

            Args:
                relay_number (int):
                    * The relay to configure.
                    * Options are:
                        * 1 or 2

                activating_input_channel (str):
                    * Specifies which input alarm activates the relay when the relay is in alarm mode
                    * Only applies if ALARM mode is chosen.
                    * Options are:
                        * A
                        * B
                        * C(1 - 5)
                        * D(1 - 5)

                alarm_relay_trigger_type (Model224RelayControlAlarm):
                    * Specifies the type of alarm that triggers the relay
                    * Only applies if ALARM mode is chosen.

        """
        self.command(f"RELAY {relay_number},2,{activating_input_channel},{alarm_relay_trigger_type}")

    def get_relay_alarm_control_parameters(self, relay_number):
        """Returns the relay alarm configuration for either of the two configurable relays. Relay must be
        configured for alarm mode to retrieve parameters.

            Args:
                relay_number (int)
                    * Specifies which relay to query
                    * Options are:
                        * 1 or 2

            Return:
                (dict):
                    {activating_input_channel: str, alarm_relay_trigger_type: Model224RelayControlAlarm}

        """

        relay_config = self.query(f"RELAY? {relay_number}").split(",")
        activating_input_channel = relay_config[1]
        alarm_relay_trigger_type = Model224RelayControlAlarm(int(relay_config[2]))
        return {'activating_input_channel': activating_input_channel,
                'alarm_relay_trigger_type': alarm_relay_trigger_type}

    def get_relay_control_mode(self, relay_number):
        """Returns the configured mode of the specified relay.

            Args:
                relay_number (int):
                    * Specifies which relay to query
                    * Options are:
                        * 1 or 2

            Returns:
                (Model224RelayControlMode):
                    * The configured mode of the relay, represented as an object of the enum type
                        Model224RelayControlMode

        """
        relay_settings = self.query("RELAY? " + str(relay_number))
        split_relay_settings = relay_settings.split(",")
        return Model224RelayControlMode(int(split_relay_settings[0]))

    def _get_identity(self):
        return self.query('*IDN?', check_errors=False).split(',')


__all__ = ['Model224', 'Model224AlarmParameters', 'Model224CurveFormat', 'Model224CurveHeader',
           'Model224CurveTemperatureCoefficients', 'Model224DiodeExcitationCurrent', 'Model224DisplayMode',
           'Model224DiodeSensorRange', 'Model224InputChannel', 'Model224InterfaceMode', 'Model224DisplayFieldUnits',
           'Model224StandardEventRegister', 'Model224InputSensorSettings', 'Model224InputSensorType', 'Model224InputSensorUnits',
           'Model224NTCRTDSensorResistanceRange', 'Model224NumberOfFields', 'Model224PlatinumRTDSensorResistanceRange',
           'Model224ReadingStatusRegister', 'Model224RelayControlAlarm', 'Model224RelayControlMode', 'Model224RemoteInterface',
           'Model224ServiceRequestRegister', 'Model224SoftCalSensorTypes', 'Model224StatusByteRegister']



"""Implements functionality unique to the Lake Shore Model 335 cryogenic temperature controller"""
from enum import IntEnum

from .temperature_controllers import TemperatureController, InstrumentException
from .temperature_controllers import RelayControlMode, RelayControlAlarm, InterfaceMode, HeaterError, \
    CurveFormat, CurveTemperatureCoefficient, BrightnessLevel, AutotuneMode, HeaterResistance, Polarity, \
    DiodeCurrent, HeaterOutputUnits, InputSensorUnits, ControlTypes, StandardEventRegister, OperationEvent, RegisterBase

Model335RelayControlMode = RelayControlMode
Model335RelayControlAlarm = RelayControlAlarm
Model335InterfaceMode = InterfaceMode
Model335HeaterError = HeaterError
Model335CurveFormat = CurveFormat
Model335CurveTemperatureCoefficient = CurveTemperatureCoefficient
Model335BrightnessLevel = BrightnessLevel
Model335AutoTuneMode = AutotuneMode
Model335HeaterResistance = HeaterResistance
Model335Polarity = Polarity
Model335DiodeCurrent = DiodeCurrent
Model335HeaterOutputUnits = HeaterOutputUnits
Model335InputSensorUnits = InputSensorUnits
Model335ControlTypes = ControlTypes

Model335StandardEventRegister = StandardEventRegister
Model335OperationEvent = OperationEvent


class Model335InputSensor(IntEnum):
    """Enumeration when "NONE" is an option for sensor input"""
    NONE = 0
    CHANNEL_A = 1
    CHANNEL_B = 2


class Model335MonitorOutUnits(IntEnum):
    """Units associated with a sensor channel"""
    KELVIN = 1
    CELSIUS = 2
    SENSOR = 3


class Model335InputSensorType(IntEnum):
    """Sensor type enumeration"""
    DISABLED = 0
    DIODE = 1
    PLATINUM_RTD = 2
    NTC_RTD = 3
    THERMOCOUPLE = 4


class Model335DiodeRange(IntEnum):
    """Diode voltage range enumeration"""
    TWO_POINT_FIVE_VOLTS = 0
    TEN_VOLTS = 1


class Model335RTDRange(IntEnum):
    """RTD resistance range enumeration"""
    TEN_OHM = 0
    THIRTY_OHM = 1
    HUNDRED_OHM = 2
    THREE_HUNDRED_OHM = 3
    ONE_THOUSAND_OHM = 4
    THREE_THOUSAND_OHM = 5
    TEN_THOUSAND_OHM = 6
    THIRTY_THOUSAND_OHM = 7
    ONE_HUNDRED_THOUSAND_OHM = 8


class Model335ThermocoupleRange(IntEnum):
    """Thermocouple range enumeration"""
    FIFTY_MILLIVOLT = 0


class Model335InputSensorSettings:
    """Class object used in the get/set_input_sensor methods"""

    def __init__(self, sensor_type, autorange_enable, compensation, units, input_range=None):
        """Constructor for the InputSensor class

            Args:
                sensor_type (Model335InputSensorType):
                    * Specifies input sensor type

                autorange_enable (bool):
                    * Specifies autoranging
                    * False = off and True = on

                compensation (bool):
                    * Specifies input compensation
                    * False = off and True = on

                units (Model335InputSensorUnits):
                    * Specifies the preferred units parameter for sensor readings and for the control setpoint

                input_range (IntEnum)
                    * Specifies input range if autorange_enable is false
                    * See IntEnum classes:
                        * Model335DiodeRange
                        * Model335RTDRange
                        * Model335ThermocoupleRange

        """

        self.sensor_type = sensor_type
        self.autorange_enable = autorange_enable
        self.compensation = compensation
        self.units = units
        self.input_range = input_range


class Model335HeaterOutType(IntEnum):
    """Heater output 2 enumeration"""
    CURRENT = 0
    VOLTAGE = 1


class Model335HeaterOutputDisplay(IntEnum):
    """Heater output display units enumeration"""
    CURRENT = 1
    POWER = 2


class Model335HeaterOutputMode(IntEnum):
    """Control loop enumeration"""
    OFF = 0
    CLOSED_LOOP = 1
    ZONE = 2
    OPEN_LOOP = 3
    MONITOR_OUT = 4
    WARMUP_SUPPLY = 5


class Model335WarmupControl(IntEnum):
    """Heater output 2 voltage mode warmup enumerations"""
    AUTO_OFF = 0
    CONTINUOUS = 1


class Model335HeaterRange(IntEnum):
    """Control loop heater range enumeration"""
    OFF = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3


class Model335ControlLoopZoneSettings:
    """Control loop configuration for a particular heater output and zone"""

    def __init__(self, upper_bound, proportional, integral, derivative, manual_output_value,
                 heater_range, channel, ramp_rate):
        """Constructor

            Args:
                upper_bound (float):
                    * Specifies the upper Setpoint boundary of this zone in kelvin

                proportional (float):
                    * Specifies the proportional gain for this zone
                    * 0.1 to 1000

                integral (float):
                    * Specifies the integral gain for this zone
                    * 0.1 to 1000

                derivative (float):
                    * Specifies the derivative gain for this zone
                    * 0 to 200 %

                manual_output_value (float):
                    * Specifies the manual output for this zone
                    * 0 to 100 %

                heater_range (Model335HeaterRange):
                    * Specifies the heater range for this zone
                    * See Model335HeaterRange IntEnum class

                channel (Model335InputSensor):
                    * See Model335InputSensor IntEnum class

                ramp_rate (float):
                    * Specifies the ramp rate for this zone
                    * 0 - 100 K/min

        """

        self.upper_bound = upper_bound
        self.proportional = proportional
        self.integral = integral
        self.derivative = derivative
        self.manual_output_value = manual_output_value
        self.heater_range = heater_range
        self.channel = channel
        self.ramp_rate = ramp_rate


class Model335DisplaySetup(IntEnum):
    """Panel display setup enumeration"""
    INPUT_A = 0
    INPUT_A_MAX_MIN = 1
    TWO_INPUT_A = 2
    INPUT_B = 3
    INPUT_B_MAX_MIN = 4
    TWO_INPUT_B = 5
    CUSTOM = 6
    TWO_LOOP = 7


class Model335HeaterVoltageRange(IntEnum):
    """Voltage mode heater enumerations"""
    VOLTAGE_OFF = 0
    VOLTAGE_ON = 1


class Model335DisplayInputChannel(IntEnum):
    """Panel display information enumeration"""
    NONE = 0
    INPUT_A = 1
    INPUT_B = 2
    SETPOINT_1 = 3
    SETPOINT_2 = 4
    OUTPUT_1 = 5
    OUTPUT_2 = 6


class Model335DisplayFieldUnits(IntEnum):
    """Panel display units enumeration"""
    KELVIN = 1
    CELSIUS = 2
    SENSOR_UNITS = 3
    MINIMUM_DATA = 4
    MAXIMUM_DATA = 5
    SENSOR_NAME = 6


class Model335StatusByteRegister(RegisterBase):
    """Class object representing the status byte register LSB to MSB"""

    bit_names = [
        "",
        "",
        "",
        "",
        "message_available_summary_bit",
        "event_status_summary_bit",
        "service_request",
        "operation_summary_bit"
    ]

    def __init__(self,
                 message_available_summary_bit,
                 event_status_summary_bit,
                 service_request,
                 operation_summary_bit):
        self.message_available_summary_bit = message_available_summary_bit
        self.event_status_summary_bit = event_status_summary_bit
        self.service_request = service_request
        self.operation_summary_bit = operation_summary_bit


class Model335ServiceRequestEnable(RegisterBase):
    """Class object representing the service request enable register LSB to MSB"""

    bit_names = [
        "",
        "",
        "",
        "",
        "message_available_summary_bit",
        "event_status_summary_bit",
        "",
        "operation_summary_bit"
    ]

    def __init__(self,
                 message_available_summary_bit,
                 event_status_summary_bit,
                 operation_summary_bit):
        self.message_available_summary_bit = message_available_summary_bit
        self.event_status_summary_bit = event_status_summary_bit
        self.operation_summary_bit = operation_summary_bit


class Model335InputReadingStatus(RegisterBase):
    """Class object representing the input status flag bits"""

    bit_names = [
        "invalid_reading",
        "",
        "",
        "",
        "temp_underrange",
        "temp_overrange",
        "sensor_units_zero",
        "sensor_units_overrange"
    ]

    def __init__(self, invalid_reading, temp_underrange, temp_overrange, sensor_units_zero, sensor_units_overrange):
        self.invalid_reading = invalid_reading
        self.temp_underrange = temp_underrange
        self.temp_overrange = temp_overrange
        self.sensor_units_zero = sensor_units_zero
        self.sensor_units_overrange = sensor_units_overrange


class Model335(TemperatureController):
    """A class object representing the Lake Shore Model 335 cryogenic temperature controller"""

    # Initiate enum types for temperature controllers
    _input_channel_enum = Model335DisplayInputChannel
    _display_units_enum = Model335DisplayFieldUnits

    # Initiate instrument specific registers
    _status_byte_register = Model335StatusByteRegister
    _service_request_enable = Model335ServiceRequestEnable

    vid_pid = [(0x1FB9, 0x0300)]

    def __init__(self,
                 baud_rate,
                 serial_number=None,
                 com_port=None,
                 timeout=2.0,
                 ip_address=None,
                 tcp_port=None,
                 **kwargs):
        # Call the parent init, then fill in values specific to the 335
        TemperatureController.__init__(self, serial_number, com_port, baud_rate, timeout, ip_address,
                                       tcp_port, **kwargs)

        # Disable emulation mode
        self._disable_emulation()

    # Alias specific temperature controller methods
    get_analog_output_percentage = TemperatureController._get_analog_output_percentage
    set_autotune = TemperatureController._set_autotune
    set_brightness = TemperatureController._set_brightness
    get_brightness = TemperatureController._get_brightness
    get_operation_condition = TemperatureController._get_operation_condition
    get_operation_event_enable = TemperatureController._get_operation_event_enable
    set_operation_event_enable = TemperatureController._set_operation_event_enable
    get_operation_event = TemperatureController._get_operation_event
    get_thermocouple_junction_temp = TemperatureController._get_thermocouple_junction_temp
    set_soft_cal_curve_dt_470 = TemperatureController._set_soft_cal_curve_dt_470
    set_soft_cal_curve_pt_100 = TemperatureController._set_soft_cal_curve_pt_100
    set_soft_cal_curve_pt_1000 = TemperatureController._set_soft_cal_curve_pt_1000
    set_diode_excitation_current = TemperatureController._set_diode_excitation_current
    get_diode_excitation_current = TemperatureController._get_diode_excitation_current
    get_tuning_control_status = TemperatureController._get_tuning_control_status
    set_filter = TemperatureController._set_filter
    get_filter = TemperatureController._get_filter

    def set_monitor_output_heater(self, channel, high_value, low_value, units=Model335MonitorOutUnits.KELVIN,
                                  polarity=Model335Polarity.UNIPOLAR):
        """Configures output 2. Use the set_heater_output_mode command to set the output mode to Monitor Out.

            Args:
                channel (Model335InputSensor):
                    * Specifies which sensor input to monitor

                high_value (float):
                    * Represents the data at which the Monitor Out reaches +100% output
                    * Entered in the units designated by the <units> argument

                low_value (float):
                    * Represents the data at which the analog output reaches -100% output if bipolar,
                    * or 0% outputif unipolar. Entered in the units designated by the <units> argument

                units (Model335MonitorOutUnits):
                    * Specifies the units on which to base the output voltage

                polarity (Model335Polarity):
                    * Specifies output voltage is unipolar or bipolar

        """
        self.command(f"ANALOG 2,{channel},{units},{high_value},{low_value},{polarity}")

    def get_monitor_output_heater(self):
        """Used to obtain all monitor out parameters for output 2.

            Return:
                (dict):
                    * See set_monitor_output_heater method arguments
                    * Keys:
                        * "channel": Model335InputSensor
                        * "units": Model335MonitorOutUnits
                        * "high_value": float
                        * "low_value": float
                        * "polarity": Model335Polarity

        """
        parameters = self.query("ANALOG? 2").split(",")
        return {"channel": Model335InputSensor(int(parameters[0])),
                "units": Model335MonitorOutUnits(int(parameters[1])),
                "high_value": float(parameters[2]),
                "low_value": float(parameters[3]),
                "polarity": Model335Polarity(int(parameters[4]))}

    def get_celsius_reading(self, channel):
        """Returns the temperature value in celsius of either channel.

            Args:
                channel (str):
                    * Selects the sensor input to query
                    * "A" or "B"

        """
        return float(self.query(f"CRDG? {channel}"))

    def set_display_setup(self, mode):
        """Sets the display mode

            Args:
                mode (Model335DisplaySetup):
                    * Specifies the front panel display mode
                    * See Model335DisplaySetup IntEnum class

        """
        self.command(f"DISPLAY {mode}")

    def get_display_setup(self):
        """Returns the display mode

            Return:
                (Model335DisplaySetup):
                    * Specifies the front panel display mode
                    * See Model335DisplaySetup IntEnum class

        """
        return Model335DisplaySetup(int(self.query("DISPLAY?")))

    def set_heater_setup_one(self, heater_resistance, max_current, output_display_mode):
        """Method to configure heater output one.

            Args:
                heater_resistance (Model335HeaterResistance):
                    * See Model335HeaterResistance IntEnum class

                max_current (float):
                    * Specifies the maximum current for the heater

                output_display_mode (Model335HeaterOutputDisplay):
                    * Specifies how the heater output is displayed
                    * See Model335HeaterOutType IntEnum class

        """
        self.command(f"HTRSET 1,0,{heater_resistance},0,{max_current},{output_display_mode}")

    def set_heater_setup_two(self, output_type, heater_resistance, max_current, display_mode):
        """Method to configure the heater output 2.

            Args:
                output_type (Model335HeaterOutType):
                    * Specifies wheter the heater output is in constant current or voltage mode
                    * See Model335HeaterOutType IntEnum class

                heater_resistance (Model335HeaterResistance):
                    * See Model335HeaterResistance IntEnum class

                max_current (float):
                    * Specifies the maximum current for the heater

                display_mode (Model335HeaterOutType):
                    * Specifies how the heater output is displayed
                    * Required only if output_type is set to CURRENT
                    * See Model335HeaterOutType IntEnum class

        """
        self.command(f"HTRSET 2,{output_type},{heater_resistance},0,{max_current},{display_mode}")

    def get_heater_setup(self, heater_output):
        """Returns the heater configuration status.

            Args:
                heater_output (int)
                    * Selects which heater output to query

            Return:
                (dict):
                    * See set_heater_setup_one/set_heater_setup_two method arguments
                    * Keys:
                        * "output_type": Model335HeaterOutType
                        * "heater_resistance": Model335HeaterResistance
                        * "max_current": float
                        * "output_display_mode": Model335HeaterOutputDisplay

        """
        heater_setup = self.query(f"HTRSET? {heater_output}").split(",")
        if int(heater_setup[2]) == 0:
            max_current = float(heater_setup[3])
        else:
            preset_currents = ["USER", 0.707, 1.0, 1.141, 1.732]
            current_index = int(heater_setup[2])
            max_current = preset_currents[current_index]

        return {"output_type": Model335HeaterOutType(int(heater_setup[0])),
                "heater_resistnace": Model335HeaterResistance(int(heater_setup[1])),
                "max_current": max_current,
                "output_display_mode": Model335HeaterOutputDisplay(int(heater_setup[4]))}

    def set_input_sensor(self, channel, sensor_parameters):
        """Sets the sensor type and associated parameters.

            Args:
                channel (str):
                    * Specifies input to configure
                    * "A" or "B"

                sensor_parameters (Model335InputSensorSettings):
                    * See Model335InputSensorSettings class

        """
        autorange_enable = bool(int(sensor_parameters.autorange_enable))
        if autorange_enable:
            input_range = 0
        else:
            input_range = sensor_parameters.input_range

        command_string = (f"INTYPE {channel},{sensor_parameters.sensor_type},{int(sensor_parameters.autorange_enable)}," +
                            f"{input_range},{int(sensor_parameters.compensation)},{sensor_parameters.units}")

        self.command(command_string)

    def get_input_sensor(self, channel):
        """Returns the sensor type and associated parameters.

            Args:
                channel (str):
                    * Specifies sensor input to configure
                    * "A" or "B"

            Return:
                (Model335InputSensorSettings):
                    * See Model335InputSensor IntEnum class

        """
        sensor_configuration = self.query(f"INTYPE? {channel}").split(",")
        input_sensor_type = Model335InputSensorType(int(sensor_configuration[0]))

        sensor_range = None
        if bool(int(sensor_configuration[1])):
            sensor_range = 0
        elif input_sensor_type == Model335InputSensorType.DISABLED:
            sensor_range = 0
        elif input_sensor_type == Model335InputSensorType.DIODE:
            sensor_range = Model335DiodeRange(int(sensor_configuration[2]))
        elif input_sensor_type in (Model335InputSensorType.PLATINUM_RTD or Model335InputSensorType.NTC_RTD):
            sensor_range = Model335RTDRange(int(sensor_configuration[2]))
        elif input_sensor_type == Model335InputSensorType.THERMOCOUPLE:
            sensor_range = Model335ThermocoupleRange(int(sensor_configuration[2]))

        return Model335InputSensorSettings(input_sensor_type, bool(int(sensor_configuration[1])),
                                           bool(int(sensor_configuration[3])),
                                           Model335InputSensorUnits(int(sensor_configuration[4])),
                                           sensor_range)

    def get_all_kelvin_reading(self):
        """Returns the temperature value in kelvin of all channels.

            Return:
                (list: float)
                    * [channel_A, channel_B]

        """
        return [float(self.query("KRDG? A")), float(self.query("KRDG? B"))]

    def set_heater_output_mode(self, output, mode, channel, powerup_enable=False):
        """Configures the heater output mode.

            Args:
                output (int):
                    * Specifies which output to configure (1 or 2)

                mode (Model335HeaterOutputMode):
                    * Member of Model335HeaterOutputMode IntEnum class
                    * Specifies the control mode

                channel (Model335InputSensor)
                    * Specifies which input to use for control

                powerup_enable (bool)
                    * Specifies whether the output remains on (True)
                    * or shuts off after power cycle (False)

        """
        command_string = f"OUTMODE {output},{mode},{channel},{int(powerup_enable)}"
        self.command(command_string)

    def get_heater_output_mode(self, output):
        """Returns the heater output mode for a given output and whether powerup is enabled.

            Args:
                output (int):
                    * Specifies which output to query (1 or 2)

            Return:
                (dict):
                    * Keys:
                        * "mode": Model335HeaterOutputMode
                        * "channel": Model335InputSensor
                        * "powerup_enable": bool

        """
        outmode = self.query(f"OUTMODE? {output}").split(",")

        return {"mode": Model335HeaterOutputMode(int(outmode[0])),
                "channel": Model335InputSensor(int(outmode[1])),
                "powerup_enable": bool(int(outmode[2]))}

    def set_output_two_polarity(self, output_polarity):
        """Sets polarity of output 2 to either unipolar or bipolar, only applicable when
        output 2 is in voltage mode.

            Args:
                output_polarity (Model335Polarity)
                    * Specifies whether output voltage is UNIPOLAR or BIPOLAR

        """
        self.command(f"POLARITY 2,{output_polarity}")

    def get_output_2_polarity(self):
        """Returns the polarity of output 2

            Return:
                (Model335Polarity)
                    * Specifies whether output is UNIPOLAR or BIPOLAR

        """
        return Model335Polarity(int(self.query("POLARITY?")))

    def set_heater_range(self, output, heater_range):
        """Sets the heater range for a particular output. The range setting has no effect if an output is in
        the off mode, and does not apply to an output in Monitor Out mode. An output in Monitor Out mode is always on.

            Args:
                output (int):
                    * Specifies which output to configure (1 or 2)

                heater_range (IntEnum):
                    * For Outputs 1 and 2 in Current mode:
                        * Model335HeaterRange IntEnum member
                    * For Output 2 in Voltage mode:
                        * Model335HeaterVoltageRange IntEnum member

        """
        self.command(f"RANGE {output},{heater_range}")

    def get_heater_range(self, output):
        """Returns the heater range for a particular output.

            Args:
                output (int):
                    * Specifies which output to configure (1 or 2)

            Return:
                heater_range (IntEnum):
                    * For Outputs 1 and 2 in Current mode:
                        * Model335HeaterRange IntEnum member
                    * For Output 2 in Voltage mode:
                        * Model335HeaterVoltageRange IntEnum member

        """
        heater_range = int(self.query(f"RANGE? {output}"))
        if output == 2:
            # Check if output 2 is in voltage mode
            output_2_heater_setup = self.query("HTRSET? 2").split(",")
            output_2_voltage_enable = bool(int(output_2_heater_setup[0]))
            if output_2_voltage_enable:
                heater_range = Model335HeaterVoltageRange(heater_range)
            else:
                heater_range = Model335HeaterRange(heater_range)
        else:
            heater_range = Model335HeaterRange(heater_range)
        return heater_range

    def all_heaters_off(self):
        """Recreates the front panel safety feature of shutting off all heaters"""

        self.command("RANGE 1,0")
        self.command("RANGE 2,0")

    def get_input_reading_status(self, channel):
        """Returns the state of the input status flag bits.

            Args:
                channel (str):
                    * Specifies which channel to query
                    * "A" or "B"

            Return:
                (InputReadingStatus):
                    * Boolean representation of each bit of the input status flag register
        """
        response = int(self.query(f"RDGST? {channel}"))
        return Model335InputReadingStatus.from_integer(response)

    def set_warmup_supply(self, control, percentage):
        """Warmup mode applies only to Output 2 in Voltage mode. The Output Type parameter
        must be configured using the set_heater_setup() method, and the Output mode and Control
        Input parameters must be configured using the set_monitor_out_parameters() method.

            Args:
                control (Model335WarmupControl):
                    * Specifies the type of control used

                percentage (float):
                    * Specifies the percentage of full scale (10 V) Monitor Out voltage to apply

        """
        # Check if output 2 is in voltage mode
        output_2_heater_setup = self.query("HTRSET? 2").split(",")
        output_2_voltage_enable = bool(int(output_2_heater_setup[0]))
        if not output_2_voltage_enable:
            raise InstrumentException("Output 2 is not configured in voltage mode")

        command_string = f"WARMUP 2,{control},{percentage}"
        self.command(command_string)

    def get_warmup_supply(self):
        """Returns the output 2 warmup supply configuration.

            Return:
                (dict):
                    * Keys:
                        * "control": Model335WarmupControl
                        * "percentage": float

        """
        warmup_supply = self.query("WARMUP? 2").split(",")
        return {"control": Model335WarmupControl(int(warmup_supply[0])),
                "percentage": float(warmup_supply[1])}

    def set_control_loop_zone_table(self, output, zone, control_loop_zone):
        """Configures the output zone parameters.

            Args:
                output (int):
                    * Specifies which heater output to configure
                    * 1 or 2

                zone (int):
                    * Specifies which zone in the table to configure
                    * 1 to 10

                control_loop_zone (ControlLoopZone):
                    * See ControlLoopZone class

        """
        command_string = (f"ZONE {output},{zone},{control_loop_zone.upper_bound},{control_loop_zone.proportional}," +
                            f"{control_loop_zone.integral},{control_loop_zone.derivative},{control_loop_zone.manual_output_value}," +
                            f"{control_loop_zone.heater_range},{control_loop_zone.channel},{control_loop_zone.ramp_rate}")
        self.command(command_string)

    def get_control_loop_zone_table(self, output, zone):
        """Returns a list of zone control parameters for a selected output and zone.

            Args:
                output (int):
                    * Specifies which heater output to query
                    * 1 or 2

                zone (int):
                    * Specifies which zone in the table to query
                    * 1 to 10

            Return:
                (Model335ControlLoopZone):
                    * See Model335ControlLoopZone class

        """
        zone_parameters = self.query(f"ZONE? {output},{zone}").split(",")
        control_loop_zone_parameters = Model335ControlLoopZoneSettings(float(zone_parameters[0]),
                                                                       float(zone_parameters[1]),
                                                                       float(zone_parameters[2]),
                                                                       float(zone_parameters[3]),
                                                                       float(zone_parameters[4]),
                                                                       Model335HeaterRange(int(zone_parameters[5])),
                                                                       Model335InputSensor(int(zone_parameters[6])),
                                                                       float(zone_parameters[7]))
        return control_loop_zone_parameters

    def _disable_emulation(self):
        """Disables emulation mode so that instrument is compatible with Python Driver."""
        self.command("EMUL 0", check_errors=False)


__all__ = ['Model335', 'Model335AutoTuneMode', 'Model335BrightnessLevel', 'Model335ControlLoopZoneSettings',
           'Model335ControlTypes', 'Model335CurveFormat', 'Model335CurveFormat', 'Model335CurveTemperatureCoefficient',
           'Model335DiodeCurrent', 'Model335DiodeRange', 'Model335DisplayFieldUnits', 'Model335DisplayInputChannel',
           'Model335DisplaySetup', 'Model335HeaterError', 'Model335HeaterOutputDisplay', 'Model335HeaterOutputMode',
           'Model335HeaterOutputUnits', 'Model335HeaterOutType', 'Model335HeaterRange', 'Model335HeaterResistance',
           'Model335HeaterVoltageRange', 'Model335InputReadingStatus', 'Model335InputSensor',
           'Model335InputSensorSettings', 'Model335InputSensorType', 'Model335InputSensorUnits',
           'Model335InterfaceMode', 'Model335MonitorOutUnits', 'Model335OperationEvent', 'Model335Polarity',
           'Model335RelayControlAlarm', 'Model335RelayControlMode', 'Model335RTDRange', 'Model335ServiceRequestEnable',
           'Model335StandardEventRegister', 'Model335StatusByteRegister', 'Model335ThermocoupleRange',
           'Model335WarmupControl']



"""Python driver for Lake Shore instruments"""
from .generic_instrument import InstrumentException
from .xip_instrument import XIPInstrumentException
from .teslameter import Teslameter, TeslameterOperationRegister, TeslameterQuestionableRegister, F41, F71
from .fast_hall_controller import FastHall, FastHallOperationRegister, FastHallQuestionableRegister, ContactCheckManualParameters,\
    ContactCheckOptimizedParameters, FastHallManualParameters, FastHallLinkParameters, FourWireParameters,\
    DCHallParameters, ResistivityManualParameters, ResistivityLinkParameters, M91
from .model_155 import PrecisionSource, PrecisionSourceOperationRegister, PrecisionSourceQuestionableRegister, Model155
from .model_121 import Model121
from .model_224 import *
from .model_240 import *
from .model_335 import *
from .model_336 import *
from .model_336 import Model336
from .model_350 import Model350
from .model_372 import *
from .model_425 import Model425
from .model_643 import Model643
from .model_648 import Model648
from .ssm_system import SSMSystem, SSMSystemQuestionableRegister, SSMSystemOperationRegister
from .ssm_system_enums import SSMSystemEnums
from .ssm_base_module import SSMSystemModuleQuestionableRegister
from .ssm_measure_module import SSMSystemMeasureModuleOperationRegister
from .ssm_source_module import SSMSystemSourceModuleOperationRegister
